////////////////////////////////////////////////////////////////
//  GJVIFS.CPP:  C++ declaration and methods for an iterated 
//               function system (IFS)
//  Garry J. Vass, Compuserve 72307,3311
//                 Bix  gjvass
//
//
//  The code below defines a C++ class and methods for 
//  encapsulating an IFS object.  The IFS decoding algorithm
//  was descibed in Byte Magazine (Jan, 1988, Barnsley).
//  Credit should go to Samuel Druker, who originally published
//  a ZORTECH 'C' version in Jan, 1989 using the Flash Graphics
//  subsystem ("CHAOS.C"), and to Marty Belles, who published 
//  a WINAPP version ("TRICHO.C") in Sep. 1989.  Thanks and a
//  tip of the hat to pioneers Druker and Belles!
//
//  My efforts focused on using C++ to encapsulate/hide most
//  of the mechanical operations and to provide a streamlined
//  way to create and view IFS designs.  I was also interested
//  in separating the IFS stuff from the graphics engine and in
//  devising a graphics engine that allowed portability between
//  TC++ and ZORTECH.  As it is not of primary interest here,
//  the methods for the graphics engine are given elsewhere.
//
//  At the time of its creation, this program compiled and
//  ran flawlessly under both TC and ZWB (appropriate config
//  information is included in the archive), the key parameters
//  are word aligned, large model, unsigned characters, and
//  286 instructions.
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
#ifndef __TURBOC__
#define ZORTECH
#endif
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
#include   <stdlib.h>
#include     <time.h>
#include     <math.h>
#include    <ctype.h>
#include    "gjvog.hpp"
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
#define FCOLOR(v)    ( v[ 0 ] )
#define BCOLOR(v)    ( v[ 1 ] )
#define XPHYSICAL(v) ( v[ 2 ] )
#define YPHYSICAL(v) ( v[ 3 ] )
#define XWORLD(v)    ( v[ 4 ] )
#define YWORLD(v)    ( v[ 5 ] )
#define XORIGIN(v)   ( v[ 6 ] )
#define YORIGIN(v)   ( v[ 7 ] )
#define ITERATIONS(v)( v[ 0 ] )
#define XLOGICAL(v)  ( v[ 1 ] )
#define YLOGICAL(v)  ( v[ 2 ] )
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
class cGJVIFR {
private:

    double a[ 4 ];
    double b[ 4 ];
    double c[ 4 ];
    double d[ 4 ];
    double e[ 4 ];
    double f[ 4 ];
    int    p[ 4 ];
    int    ivals[ 10 ];
    double dvals[ 10 ];

public:

    // Simple constructor
    cGJVIFR() {
        FCOLOR(ivals) = 3;
        BCOLOR(ivals) = 1;
        ITERATIONS(dvals) = 10000.0;
    }

    // Constructor to set world and colors
    cGJVIFR( int xw, int yw, int xo, int yo, int fore, int back ) {
        set_world( xw, yw, xo, yo );
        set_colors( fore, back );
        ITERATIONS(dvals) = 10000.0;
    }

    // Reset iteration limit
    void set_iterations( double newits ) {
        ITERATIONS(dvals) = newits;
    }

    // Initialize one of the six ifs vectors
    int set_vect( char which, double a0, double a1, double a2, double a3 ) {
        double *dptr = a;
        switch( toupper( which ) ) {
                case 'A': dptr = a; break;
                case 'B': dptr = b; break;
                case 'C': dptr = c; break;
                case 'D': dptr = d; break;
                case 'E': dptr = e; break;
                case 'F': dptr = f; break;
        }
        dptr[ 0 ] = a0;
        dptr[ 1 ] = a1;
        dptr[ 2 ] = a2;
        dptr[ 3 ] = a3;
        return( 1 );
    }

    // Initialize the probability vector
    int set_vect( int a0, int a1, int a2, int a3 ) {
        int *iptr = p;
        iptr[ 0 ] = a0;
        iptr[ 1 ] = a1;
        iptr[ 2 ] = a2;
        iptr[ 3 ] = a3;
        return( 1 );
    }

    // Set world values
    int set_world( int xs, int ys, int xo, int yo ) {
        XWORLD(ivals)  = xs;
        YWORLD(ivals)  = ys;
        XORIGIN(ivals) = xo;
        YORIGIN(ivals) = yo;
        return( 1 );
    }

    // Set plotting colors
    void set_colors( int fore, int back ) {
        FCOLOR(ivals) = fore;
        BCOLOR(ivals) = back;
    }

    // Select next transform according to probability
    int select_index( int prob ) {
        for( int j = 0; j < 3; j++ )
             if( prob < p[ j ] )
                 return( j );
        return( 3 );
    }

    // Calculate the IFS transform
    int calc( int index ) {
        double temp = ( a[ index ] * XLOGICAL(dvals) + b[ index ] * YLOGICAL(dvals) + e[ index ] );
        YLOGICAL(dvals) = ( c[ index ] * XLOGICAL(dvals) + d[ index ] * YLOGICAL(dvals) + f[ index ] );
        XLOGICAL(dvals) = temp;
        return( index );
    }

    //  Plot the IFS
    void go( ) {
        // instantiate an EGA/VGA graphics mode
        cGJVGR g( 16 );
        // paint the background
        g.wipe( BCOLOR(ivals) );
        for( double i = 1; i <= ITERATIONS(dvals); i++ ) {
             calc( select_index( RANDMAX( p[ 3 ] ) ) );
             XPHYSICAL(ivals) = XLOGICAL(dvals) * XWORLD(ivals) + XORIGIN(ivals);
             YPHYSICAL(ivals) = YLOGICAL(dvals) * YWORLD(ivals) + YORIGIN(ivals);
             XPHYSICAL(ivals) = ( XPHYSICAL(ivals) + ( g.right() / 2 ) );
             YPHYSICAL(ivals) = ( ( g.bottom() / 2 ) -
                                ( ( ( g.bottom() / 4 ) *
                                YPHYSICAL(ivals) ) >> 7 ) );
             g.point( XPHYSICAL(ivals), YPHYSICAL(ivals), FCOLOR(ivals) );
        }
    }
}; // end of class cGJVIFS
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//  The code below here implements a benchmark and exercising
//  of the cGJVIFS class.  Two easily recognizable IFS structures
//  are shown, the TREE and the FERN.
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//  This function sets up and displays the fern classic
int ifs_fern()
{
    // Do the traditional IFS fern.
    cGJVIFR Fern( 300, 300, -50, -180, 10, 1 );
    Fern.set_vect( 'A',  0.00,  0.20, -0.15,   0.85 );
    Fern.set_vect( 'B',  0.00, -0.26,  0.28,   0.04 );
    Fern.set_vect( 'C',  0.00,  0.23,  0.26,  -0.04 );
    Fern.set_vect( 'D',  0.16,  0.22,  0.24,   0.85 );
    Fern.set_vect( 'E',  0.00,  0.00,  0.00,   0.00 );
    Fern.set_vect( 'F',  0.00,  0.20,  0.20,   0.20 );
    Fern.set_vect(       328,   2621,  4915,   32676 );
    Fern.go();
    getch();
    // Play games with it.  Step into the bushes.
    Fern.set_world( 2000, 2300, 0, -180 );
    Fern.set_iterations( 30000.0 );
    Fern.go();
    return( getch() );
}
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
// This function sets up and displays the tree classic
int ifs_tree()
{
    cGJVIFR Tree( 1200, 1200, 0, -300, 63, 1 );
    Tree.set_vect( 'A',  0.0,  0.1,   0.42,  0.42 );
    Tree.set_vect( 'B',  0.0,  0.0,  -0.42,  0.42 );
    Tree.set_vect( 'C',  0.0,  0.0,   0.42, -0.42 );
    Tree.set_vect( 'D',  0.5,  0.1,   0.42,  0.42 );
    Tree.set_vect( 'E',  0.0,  0.0,   0.00,  0.00 );
    Tree.set_vect( 'F',  0.0,  0.2,   0.20,  0.20 );
    Tree.set_vect(       1638, 6553,  19660, 32767 );
    Tree.go();
    return( getch() );
}
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
int main()
{
    ifs_tree();   // do the tree ifs
    ifs_fern();   // do the fern ifs
    cGJVGR( 3 );  // instantiate a text mode
}
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//  End of gjvifs.cpp
//  Garry J. Vass, Oct., 1990
//  72307,3311
////////////////////////////////////////////////////////////////
